using System;
using System.ComponentModel;
using System.Drawing;
using System.Web;
using System.Web.UI;
using System.Web.UI.Design;
using System.Web.UI.WebControls;
using RepPP;

namespace RepPP.Toolkit.Web {

	/// <summary>Position of the buttons</summary>
	public enum ButtonPositionE {
	    /// <summary>Appears in rows containing items</summary>
	    Item                =   1,
	    /// <summary>Appears in the header</summary>
	    Header              =   2,
	    /// <summary>Appears in the footer</summary>
	    Footer              =   4,
	    /// <summary>Appears in items and header</summary>
	    ItemAndHeader       =   3,
	    /// <summary>Appears in items and footer</summary>
	    ItemAndFooter       =   5,
	    /// <summary>Appears in header and footer</summary>
	    HeaderAndFooter     =   6,
	    /// <summary>Appears everywhere</summary>
	    Everywhere          =   7
	};
    //********************************************************************************
    //
	/// <summary>
	/// Implement a button or hyperlink which can appear in a datagrid which are calling a 
	/// javascript function instead of doing a callback.
	/// </summary>
	//
    //********************************************************************************
	public class ButtonColumnNP : DataGridColumn {
	    /// <summary>Position of the button</summary>
	    private ButtonPositionE     m_eButtonPos;
	    /// <summary>Javascript command to include for header</summary>
	    private string              m_strCmdHeader;
	    /// <summary>Javascript command to include for item</summary>
	    private string              m_strCmdItem;
	    /// <summary>Javascript command to include for footer</summary>
	    private string              m_strCmdFooter;
	    /// <summary>Field descriptor for data binding</summary>
        private PropertyDescriptor  m_propDescField;

        //********************************************************************************
        //
        /// <summary>
        /// ctor:           Class constructor
        /// </summary>
	    //
        //********************************************************************************
		public ButtonColumnNP() {
		    m_eButtonPos = ButtonPositionE.Item;
		}

        //********************************************************************************
        //
        /// <summary>
        /// FormatDataTextValue:        Format the value
        /// </summary>
        /// <param name="dataTextValue">Value to format</param>
        /// <returns>
        /// Formatted value
        /// </returns>
        //
        //********************************************************************************
        protected virtual string FormatDataTextValue(object dataTextValue) {
            string  strRetVal;
            string  strFormatMask;
            
            strRetVal = string.Empty;
            if (dataTextValue != null && dataTextValue != DBNull.Value) {
                strFormatMask = DataTextFormatString;
                if (strFormatMask.Length == 0) {
                    strRetVal = dataTextValue.ToString();
                } else {
                    strRetVal = string.Format(strFormatMask, dataTextValue);
                }
            }
            return(strRetVal);
        }

        //********************************************************************************
        //
        /// <summary>
        /// Initialize:     Initialize the control
        /// </summary>
        //
        //********************************************************************************
        public override void Initialize() {
            base.Initialize();
            this.m_propDescField = null;
        }

        //********************************************************************************
        //
		/// <summary>
		/// CreateButton:       Creates a button
		/// </summary>
		/// <param name="ePos"> Position of the button</param>
		/// <returns>
		/// Control to create
		/// </returns>
        //
        //********************************************************************************
		protected virtual WebControl CreateButton(ButtonPositionE ePos) {
            WebControl              ctlRetVal;
            DataGridLinkButtonNP    linkBut;
            ButtonNP                but;
            string                  strText;
            string                  strCmd;

		    switch(ePos) {
		    case ButtonPositionE.Header:
		        strText = HeaderText;
		        strCmd  = HeaderJavaScriptCommand;
		        break;
		    case ButtonPositionE.Item:
		        strText = Text;
		        strCmd  = JavaScriptCommand;
		        break;
		    case ButtonPositionE.Footer:
		        strText = FooterText;
		        strCmd  = FooterJavaScriptCommand;
		        break;
		    default:
		        throw new System.ApplicationException("Unknown position");
		    }
		    if (ButtonType == ButtonColumnType.LinkButton) {
                linkBut                     = new DataGridLinkButtonNP();
                linkBut.Text                = strText;
                linkBut.JavaScriptCommand   = strCmd;
                ctlRetVal                   = linkBut;
		    } else {
		        but                         = new ButtonNP();
		        but.Text                    = strText;
		        but.JavaScriptCommand       = strCmd;
		        ctlRetVal                   = but;
		    }
		    return(ctlRetVal);
		}

        //********************************************************************************
        //
        /// <summary>
        /// InitializeCell:             Initialize the cell in the table
        /// </summary>
        /// <param name="cell">         Cell to initialize</param>
        /// <param name="iColIndex">    Column index</param>
        /// <param name="eItemType">    Type of cell</param>
	    //
        //********************************************************************************
        public override void InitializeCell(TableCell cell, int iColIndex, ListItemType eItemType) {
            WebControl  ctl = null;
            
            base.InitializeCell(cell, iColIndex, eItemType);
            switch(eItemType) {
            case ListItemType.Header:
                if ((ButtonPosition & ButtonPositionE.Header) == ButtonPositionE.Header) {
                    ctl = CreateButton(ButtonPositionE.Header);
                    cell.Controls.Add(ctl);
                }
                break;
            case ListItemType.Footer:
                if ((ButtonPosition & ButtonPositionE.Footer) == ButtonPositionE.Footer) {
                    ctl = CreateButton(ButtonPositionE.Footer);
                    cell.Controls.Add(ctl);
                }
                break;
            case ListItemType.Item:
            case ListItemType.AlternatingItem:
            case ListItemType.SelectedItem:
                if ((ButtonPosition & ButtonPositionE.Item) == ButtonPositionE.Item) {
                    ctl = CreateButton(ButtonPositionE.Item);
                    cell.Controls.Add(ctl);
                }
                break;
            }
            if (ctl != null && DataTextField.Length != 0) {
                ctl.DataBinding += new EventHandler(this.OnDataBindColumn);
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// OnDataBindColumn:       Called when the binding occurs on a column
        /// </summary>
        /// <param name="sender">   Sender object</param>
        /// <param name="e">        Event argument</param>
        //
        //********************************************************************************
        private void OnDataBindColumn(object sender, EventArgs e) {
            string          strDataTextField;
            Control         ctl;
            DataGridItem    item;
            object          objDataItem;
            object          objValue;
            string          strFormattedValue;

            ctl         = (Control)sender;
            item        = (DataGridItem)ctl.NamingContainer;
            objDataItem = item.DataItem;
            if (m_propDescField == null) {
                strDataTextField = DataTextField;
                m_propDescField  = TypeDescriptor.GetProperties(objDataItem).Find(strDataTextField, true);
                if (m_propDescField == null && !base.DesignMode) {
                    throw new HttpException(String.Format("Field not found {0}", strDataTextField));
                }
            }
            if (m_propDescField != null) {
                objValue          = m_propDescField.GetValue(objDataItem);
                strFormattedValue = FormatDataTextValue(objValue);
            } else {
                strFormattedValue = "Sample Databound Text";
            }
            if (ctl is LinkButtonNP) {
                ((LinkButtonNP)ctl).Text = strFormattedValue;
            } else {
                ((ButtonNP)ctl).Text     = strFormattedValue;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ButtonType:     Type of button
        /// </summary>
        //
        //********************************************************************************
        [Description("The type of button contained within the column.")]
        [Category("Appearance")]
        [DefaultValue(0)]
        public virtual ButtonColumnType ButtonType {
            get {
                ButtonColumnType    eRetVal;
                object              obj;
                
                obj = ViewState["ButtonType"];
                if (obj != null)            {
                    eRetVal = (ButtonColumnType)obj;
                } else {
                    eRetVal = ButtonColumnType.LinkButton;
                }
                return(eRetVal);
            }
            set {
                if (value < ButtonColumnType.LinkButton || value > ButtonColumnType.PushButton) {
                    throw new ArgumentOutOfRangeException("value");
                }
                ViewState["ButtonType"] = value;
                OnColumnChanged();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// CommandName:        Name of the command
        /// </summary>
        //
        //********************************************************************************
        [Description("The command associated with the button.")]
        [DefaultValue("")]
        [Category("Behavior")]
        public virtual string CommandName {
            get {
                string  strRetVal;
                
                strRetVal = ViewState["CommandName"] as string;
                if (strRetVal == null) {
                    strRetVal = string.Empty;
                }
                return(strRetVal);
            }
            set {
                ViewState["CommandName"] = value;
                OnColumnChanged();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DataTextField:      Data text field
        /// </summary>
        //
        //********************************************************************************
        [Category("Data")]
        [DefaultValue("")]
        [Description("The field bound to the text property of the button.")]
        public virtual string DataTextField {
            get {
                string  strRetVal;
                
                strRetVal = ViewState["DataTextField"] as string;
                if (strRetVal == null) {
                    strRetVal = string.Empty;
                }
                return(strRetVal);
            }
            set {
                ViewState["DataTextField"] = value;
                OnColumnChanged();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// DataTextFormatString:   The formatting applied to the value bound to the Text property
        /// </summary>
        //
        //********************************************************************************
        [Category("Data")]
        [DefaultValue("")]
        [Description("The formatting applied to the value bound to the Text property.")]
        public virtual string DataTextFormatString {
            get {
                string  strRetVal;
                
                strRetVal = ViewState["DataTextFormatString"] as string;
                if (strRetVal == null) {
                    strRetVal = string.Empty;
                }
                return(strRetVal);
            }
            set {
                ViewState["DataTextFormatString"] = value;
                OnColumnChanged();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Text:                       Text of the link
        /// </summary>
	    //
        //********************************************************************************
        [Category("Appearance")]
        [Bindable(true)]
        [PersistenceMode(PersistenceMode.InnerDefaultProperty)]
        [DefaultValue("")]
        [Description("The text to be show for the button.")]
        public virtual string Text {
            get {
                string  strRetVal;
                object  obj;
                
                obj = ViewState["Text"];
                strRetVal = (obj != null) ? (string)obj : string.Empty;
                return(strRetVal);
            }
            set {
                ViewState["Text"] = value;
                OnColumnChanged();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// HeaderJavaScriptCommand:    Javascript command
        /// </summary>
        //
        //********************************************************************************
        [Category("Behavior")]
        [Bindable(true)]
        [PersistenceMode(PersistenceMode.InnerProperty)]
        [DefaultValue("")]
        [Description("The javascript command to attach to the header.")]
        public virtual string HeaderJavaScriptCommand {
            get {
                return(m_strCmdHeader);
            }
            set {
                m_strCmdHeader = (value == null) ? string.Empty : value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// FooterJavaScriptCommand:    Javascript command
        /// </summary>
        //
        //********************************************************************************
        [Category("Behavior")]
        [Bindable(true)]
        [PersistenceMode(PersistenceMode.InnerProperty)]
        [DefaultValue("")]
        [Description("The javascript command to attach to the footer.")]
        public virtual string FooterJavaScriptCommand {
            get {
                return(m_strCmdFooter);
            }
            set {
                m_strCmdFooter = (value == null) ? string.Empty : value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// JavaScriptCommand:      Javascript command for normal item
        /// </summary>
        //
        //********************************************************************************
        [Category("Behavior")]
        [Bindable(true)]
        [PersistenceMode(PersistenceMode.InnerProperty)]
        [DefaultValue("")]
        [Description("The javascript command to attach to the line items")]
        public virtual string JavaScriptCommand {
            get {
                return(m_strCmdItem);
            }
            set {
                m_strCmdItem = (value == null) ? string.Empty : value;
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// ButtonPosition: Position of the button
        /// </summary>
        //
        //********************************************************************************
        [Description("Position of the button in the grid.")]
        [Category("Appearance")]
        [DefaultValue(ButtonPositionE.Item)]
        public virtual ButtonPositionE ButtonPosition {
            get {
                return(m_eButtonPos);
            }
            set {
                if (value < ButtonPositionE.Item || value > ButtonPositionE.Everywhere) {
                    throw new ArgumentOutOfRangeException("value");
                }
                m_eButtonPos = value;
                OnColumnChanged();
            }
        }

        //********************************************************************************
        //
        /// <summary>
        /// Implement a datagrid link button which is not posting back when pushed
        /// </summary>
	    //
        //********************************************************************************
        private class DataGridLinkButtonNP : LinkButtonNP {

            //********************************************************************************
            //
            /// <summary>
            /// ctor:           Class constructor
            /// </summary>
            //
            //********************************************************************************
            public DataGridLinkButtonNP() {
            }

            //********************************************************************************
            //
            /// <summary>
            /// SetForeColor:       Set the foreground color using the parent setting
            /// </summary>
            //
            //********************************************************************************
            private void SetForeColor() {
                Control ctlParent;
                Color   color;
                
                if (ForeColor == Color.Empty) {
                    ctlParent = this;
                    for (int iIndex = 0; iIndex < 3; iIndex++) {
                        ctlParent = ctlParent.Parent;
                        color     = ((WebControl)ctlParent).ForeColor;
                        if (color != Color.Empty) {
                            ForeColor = color;
                            break;
                        }
                    }
                }
            }

            //********************************************************************************
            //
            /// <summary>
            /// Render:                 Render the control to the stream
            /// </summary>
            /// <param name="writer">   Stream</param>
            //
            //********************************************************************************
            protected override void Render(HtmlTextWriter writer) {
                SetForeColor();
                base.Render (writer);
            }
        } // class DataGridLinkButton
	} // Class ColumnButtonNP
}
